/**
  ASCII Table Utility

  A simple utility that uses ncurses to display the ASCII table and lets you
  manually output ASCII characters one byte at a time. Use arrow keys or
  h/j/k/l to move, 'enter' to add a character to the output, and 'q' or escape
  to quit and print the output.

  Copyright 2018 Bruce Hill

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include <curses.h>
#include <ncurses.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

// The descriptive names of the ASCII characters
const char *NAMES[128][2] = {
    {"NUL", "null"}, // 0
    {"SOH", "start of heading"}, // 1
    {"STX", "start of text"}, // 2
    {"ETX", "end of text"}, // 3
    {"EOT", "end of transmission"}, // 4
    {"ENQ", "enquiry"}, // 5
    {"ACK", "acknowledge"}, // 6
    {"BEL", "bell"}, // 7
    {"BS", "backspace"}, // 8
    {"TAB", "horizontal tab"}, // 9
    {"LF", "NL line feed, new line"}, // 10
    {"VT", "vertical tab"}, // 11
    {"FF", "NP form feed, new page"}, // 12
    {"CR", "carriage return"}, // 13
    {"SO", "shift out"}, // 14
    {"SI", "shift in"}, // 15
    {"DLE", "data link escape"}, // 16
    {"DC1", "device control 1"}, // 17
    {"DC2", "device control 2"}, // 18
    {"DC3", "device control 3"}, // 19
    {"DC4", "device control 4"}, // 20
    {"NAK", "negative acknowledge"}, // 21
    {"SYN", "synchonous idle"}, // 22
    {"ETB", "end of transmission block"}, // 23
    {"CAN", "cancel"}, // 24
    {"EM", "end of medium"}, // 25
    {"SUB", "substitute"}, // 26
    {"ESC", "escape"}, // 27
    {"FS", "file separator"}, // 28
    {"GS", "group separator"}, // 29
    {"RS", "record separator"}, // 30
    {"US", "unit separator"}, // 31
    {" ", "space"}, // 32
    [127] = {"DEL", "delete"}
};

static int new_colorpair(int fg, int bg)
{
    static int next_colornum = 1;
    init_pair(next_colornum, fg, bg);
    return COLOR_PAIR(next_colornum++);
}

int DECIMAL_COLORS, HEX_COLORS, CHAR_COLORS, DESCRIPTION_COLORS, OUTPUT_CHAR_COLORS,
    OUTPUT_ESCAPE_COLORS, OUTPUT_LABEL_COLORS;

const int COL_WIDTH = 42;
int W = 0;
int H = 0;

static inline int min(int a, int b) {
    return a < b ? a : b;
}

static inline int max(int a, int b) {
    return a >= b ? a : b;
}

// Shift the cursor
static inline void shift(int dy, int dx) {
    int x, y;
    getyx(stdscr, y, x);
    move(y+dy, x+dx);
}

#define printwattrs(_attrs,fmt, ...) \
    { char _buf[COL_WIDTH+1];\
    chtype _bufch[COL_WIDTH+1];\
    sprintf(_buf, fmt, __VA_ARGS__);\
    for (int i = 0; i == 0 || _buf[i-1]; i++) _bufch[i] = _buf[i] | _attrs;\
    addchstr(_bufch);\
    shift(0, strlen(_buf));}


// Draw the information for one character
static inline void draw_entry(int i, int scrollx, int attrs)
{
    int x = (i / (H-1)) * COL_WIDTH - scrollx;
    if (x > W || x < 0) return;
    int y = i % (H-1);
    move(y, x);
    printwattrs(DECIMAL_COLORS | attrs, "%3d ", i);
    printwattrs(HEX_COLORS | attrs, " 0x%02X ", i);
    if (NAMES[i][0]) {
        printwattrs(CHAR_COLORS | attrs, " %-3s ", NAMES[i][0]);
    } else {
        printwattrs(CHAR_COLORS | attrs, " %c   ", i);
    }
    if (NAMES[i][1]) {
        printwattrs(DESCRIPTION_COLORS | A_BOLD | attrs, " %-25s", NAMES[i][1]);
    } else if (('a' <= i && i <= 'z') || ('A' <= i && i <= 'Z')) {
        printwattrs(DESCRIPTION_COLORS | attrs, " The letter %c             ", i);
    } else if ('0' <= i && i <= '9') {
        printwattrs(DESCRIPTION_COLORS | attrs, " The number %c             ", i);
    } else {
        printwattrs(DESCRIPTION_COLORS | attrs, " The %c character          ", i);
    }
}

// Redraw the parts of the ASCII table necessary for the selected item
static void redraw(int selected)
{
    static int last_selected = -1, last_scrollx = -1;
    int selectedx = (selected / (H-1)) * COL_WIDTH;
    int scrollx = (selectedx + COL_WIDTH < W) ? 0 : (((selectedx + COL_WIDTH) - W)/COL_WIDTH+1)*COL_WIDTH;
    if (selected == last_selected) return;
    if (last_selected == -1 || last_scrollx != scrollx) {
        // Redraw everything
        erase();
        for (int i = 0; i < 128; i++) {
            draw_entry(i, scrollx, i == selected ? A_REVERSE : 0);
        }
        last_scrollx = scrollx;
    } else {
        draw_entry(selected, scrollx, A_REVERSE);
        draw_entry(last_selected, scrollx, 0);
    }
	refresh();
    last_selected = selected;
}

void draw_output(char *buf, int size)
{
    attron(OUTPUT_LABEL_COLORS);
    mvprintw(H-1, 0, " Output:");
    attroff(OUTPUT_LABEL_COLORS);
    for (int i = 0; i < size; i++) {
        char c = buf[i];
        if (' ' <= c && c <= '~') { // printable range
            printwattrs(OUTPUT_CHAR_COLORS, "%c", c);
        } else {
            printwattrs(OUTPUT_ESCAPE_COLORS, "\\x%02X", c);
        }
    }
}

int main(int argc, char *argv[])
{	
    if (argc >= 2) {
        for (int i = 1; i < argc; i++) {
            putc(atoi(argv[i]), stdout);
        }
        return 0;
    }

    size_t buf_size = 256, buf_i = 0;
    char *outbuf = calloc(buf_size, sizeof(char));

    SCREEN *screen = newterm(NULL, stderr, stdin);
    set_term(screen);
    set_escdelay(5);
    W = getmaxx(stdscr);
    H = getmaxy(stdscr);
    noecho();
    keypad(stdscr, 1);
    curs_set(0);
    start_color();
	DECIMAL_COLORS = new_colorpair(COLOR_YELLOW, COLOR_BLACK) | A_BOLD;
	HEX_COLORS = new_colorpair(COLOR_GREEN, COLOR_BLACK) | A_BOLD;
	CHAR_COLORS = new_colorpair(COLOR_MAGENTA, COLOR_BLACK) | A_BOLD;
	DESCRIPTION_COLORS = new_colorpair(COLOR_WHITE, COLOR_BLACK);
	OUTPUT_CHAR_COLORS = new_colorpair(COLOR_BLACK, COLOR_WHITE) | A_BOLD;
	OUTPUT_ESCAPE_COLORS = new_colorpair(COLOR_RED, COLOR_WHITE) | A_BOLD;
    OUTPUT_LABEL_COLORS = new_colorpair(COLOR_BLACK, COLOR_YELLOW);

    int selected = 0;
    while (1) {
        // Bar at the bottom of the screen showing the output text
        redraw(selected);
        draw_output(outbuf, buf_i);
        int key = getch();
        switch (key) {
            case 'j': case KEY_DOWN:
                selected++;
                break;

            case 'J':
                selected += 10;
                break;

            case 'k': case KEY_UP:
                selected--;
                break;

            case 'K':
                selected -= 10;
                break;

            case 'l': case 'L': case KEY_RIGHT: case KEY_SRIGHT:
                selected += H-1;
                break;

            case 'h': case 'H': case KEY_LEFT: case KEY_SLEFT:
                selected -= H-1;
                break;

            case KEY_RESIZE:
                W = getmaxx(stdscr);
                H = getmaxy(stdscr);
                clear();
                break;

            case KEY_ENTER: case '\n':
                outbuf[buf_i++] = (char)selected;
                if (buf_i >= buf_size) {
                    buf_size *= 2;
                    outbuf = realloc(outbuf, buf_size);
                }
                break;

            case 'q': case 'Q': case KEY_CANCEL: case KEY_CLOSE: case KEY_EXIT: case 27:
                endwin();
                write(1, outbuf, buf_i);
                return 0;
        }
        if (selected < 0) selected = 0;
        if (selected >= 128) selected = 127;
    }

	endwin();
	return 0;
}
